
import flowws
from flowws import Argument as Arg
import tensorflow as tf
from tensorflow import keras

from .internal import HUGE_FLOAT, PairwiseVectorDifference, \
    PairwiseVectorDifferenceSum, VectorAttention, Vector2VectorAttention

class CoarseGrainAttention(Vector2VectorAttention):
    def build(self, input_shape):
        v_shape = input_shape[1]
        result = super().build(input_shape[:-1])

        if self.join_fun == 'concat':
            # always joining neighborhood values and invariant values
            stdev = tf.sqrt(2./3/v_shape[-1])
            self.join_kernels.append(self.add_weight(
                name='join_kernel_{}'.format(3), shape=(v_shape[-1], v_shape[-1]),
                initializer=keras.initializers.RandomNormal(stddev=stdev)
            ))
        return result

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return
        (r_mask, v_mask, cv_mask) = mask
        return cv_mask

    def _expand_products(self, positions, values):
        (bcast, invars, covars, vs) = super()._expand_products(positions, values)
        for idx in bcast:
            idx.insert(-1 - self.rank, None)

        invars = tf.expand_dims(invars, -2 - self.rank)
        covars = tf.expand_dims(covars, -2 - self.rank)
        new_vs = [tf.expand_dims(v, -2 - self.rank) for v in vs]
        return bcast, invars, covars, new_vs

    def _intermediates(self, inputs, mask=None):
        (positions, values, child_values) = inputs
        (broadcast_indices, invariants, covariants, expanded_values) = \
            self._expand_products(positions, values)
        neighborhood_values = self.merge_fun_(*expanded_values)
        invar_values = self.value_net(invariants)

        swap_i = -self.rank - 1
        swap_j = swap_i - 1
        child_expand_indices = list(broadcast_indices[-1])
        child_expand_indices[swap_i], child_expand_indices[swap_j] = \
            child_expand_indices[swap_j], child_expand_indices[swap_i]
        child_values = child_values[child_expand_indices]

        joined_values = self.join_fun_(child_values, invar_values, neighborhood_values)

        scales = self.scale_net(joined_values)
        scores = self.score_net(joined_values)
        old_shape = tf.shape(scores)

        if mask is not None:
            (position_mask, value_mask, child_value_mask) = mask
            if position_mask is not None:
                position_mask = tf.expand_dims(position_mask, -1)
                position_mask = tf.reduce_all([position_mask[idx] for idx in broadcast_indices[:-1]], axis=0)
            else:
                position_mask = True
            if value_mask is not None:
                value_mask = tf.expand_dims(value_mask, -1)
                value_mask = tf.reduce_all([value_mask[idx] for idx in broadcast_indices[:-1]], axis=0)
            else:
                value_mask = True
            product_mask = tf.logical_and(position_mask, value_mask)
            scores = tf.where(product_mask, scores, -HUGE_FLOAT)

        if self.reduce:
            dims = -(self.rank + 1)
            reduce_axes = tuple(-i - 2 for i in range(self.rank))
        else:
            dims = -self.rank
            reduce_axes = tuple(-i - 2 for i in range(self.rank - 1))

        shape = tf.concat([old_shape[:dims], tf.math.reduce_prod(old_shape[dims:], keepdims=True)], -1)
        scores = tf.reshape(scores, shape)
        attention = tf.reshape(tf.nn.softmax(scores), old_shape)
        output = tf.reduce_sum(attention*covariants*scales, reduce_axes)

        return dict(attention=attention, output=output, invariants=invariants)

@flowws.add_stage_arguments
class PDBInverseCoarseGrain(flowws.Stage):
    """Build a geometric attention network for a coarse-grain backmapping task.

    This module specifies the architecture of a network to produce
    atomic coordinates from a set of coarse-grained beads.

    """

    ARGS = [
        Arg('rank', None, int, 2,
            help='Degree of correlations (n-vectors) to consider'),
        Arg('n_dim', '-n', int, 32,
            help='Working dimensionality of point representations'),
        Arg('dilation', None, float, 2,
            help='Working dimension dilation factor for MLP components'),
        Arg('merge_fun', '-m', str, 'concat',
            help='Method to merge point representations'),
        Arg('join_fun', '-j', str, 'concat',
            help='Method to join invariant and point representations'),
        Arg('n_blocks_coarse', None, int, 2,
            help='Number of deep blocks to use in the coarse-grain space'),
        Arg('n_blocks_fine', None, int, 2,
            help='Number of deep blocks to use in the coarse-grain space'),
        Arg('block_nonlinearity', None, bool, True,
            help='If True, add a nonlinearity to the end of each block'),
        Arg('residual', '-r', bool, True,
            help='If True, use residual connections within blocks'),
        Arg('activation', '-a', str, 'relu',
            help='Activation function to use inside the network'),
        Arg('attention_vector_inputs', None, bool, False,
            help='Use input vectors for vector-vector attention'),
        Arg('attention_learn_projection', None, bool, False,
            help='Use learned projection weights for vector-vector attention'),
    ]

    def run(self, scope, storage):
        rank = self.arguments['rank']
        n_dim = self.arguments['n_dim']
        merge_fun = self.arguments['merge_fun']
        join_fun = self.arguments['join_fun']

        train_data = scope['train_generator']
        sample_batch = next(train_data)

        x_in = keras.layers.Input(sample_batch[0][0].shape[1:], name='rij')
        v_in = keras.layers.Input(sample_batch[0][1].shape[1:], name='tij')
        cv_in = keras.layers.Input(sample_batch[0][2].shape[1:], name='child_t')

        cv_emb = keras.layers.Embedding(len(scope['child_type_names']), n_dim, mask_zero=True)(cv_in)

        dilation_dim = round(n_dim*self.arguments['dilation'])

        def make_scorefun():
            layers = [keras.layers.Dense(dilation_dim)]

            layers.append(keras.layers.Activation(self.arguments['activation']))

            layers.append(keras.layers.Dense(1))
            return keras.models.Sequential(layers)

        def make_valuefun(dim):
            layers = [keras.layers.Dense(dilation_dim)]
            layers.append(keras.layers.LayerNormalization())

            layers.append(keras.layers.Activation(self.arguments['activation']))

            layers.append(keras.layers.Dense(dim))
            return keras.models.Sequential(layers)

        def make_block(last):
            residual_in = last
            last = VectorAttention(
                make_scorefun(), make_valuefun(n_dim), False, rank=rank,
                join_fun=join_fun,
                merge_fun=merge_fun)([x_in, last])

            if self.arguments['block_nonlinearity']:
                last = make_valuefun(n_dim)(last)

            if self.arguments['residual']:
                last = last + residual_in

            return last

        def make_vector_block(vec):
            residual_in = vec

            vec = PairwiseVectorDifference()(vec)
            (vec, ivs, att) = Vector2VectorAttention(
                make_scorefun(), make_valuefun(n_dim), make_valuefun(1), True, rank=rank,
                join_fun=join_fun, merge_fun=merge_fun,
                use_input_vectors=self.arguments['attention_vector_inputs'],
                learn_vector_projection=self.arguments['attention_learn_projection'])(
                    [vec, delta_v], return_invariants=True, return_attention=True)

            if self.arguments['residual']:
                vec = residual_in + vec

            return vec

        last = keras.layers.Dense(n_dim)(v_in)
        for _ in range(self.arguments['n_blocks_coarse']):
            last = make_block(last)

        (vec, ivs, att) = CoarseGrainAttention(
            make_scorefun(), make_valuefun(n_dim), make_valuefun(1), True, name='final_attention',
            rank=1,
            join_fun=join_fun,
            merge_fun=merge_fun)(
            [x_in, last, cv_emb], return_invariants=True, return_attention=True)

        delta_v = PairwiseVectorDifferenceSum()(cv_emb)
        delta_v = keras.layers.Dense(n_dim)(delta_v)

        for _ in range(self.arguments['n_blocks_fine']):
            vec = make_vector_block(vec)

        scope['input_symbol'] = [x_in, v_in, cv_in]
        scope['output'] = vec
        scope['loss'] = 'mse'
        scope['attention_model'] = keras.models.Model([x_in, v_in, cv_in], att)
        scope['invariant_model'] = keras.models.Model([x_in, v_in, cv_in], ivs)



import collections
import itertools

import tensorflow as tf
from tensorflow import keras

HUGE_FLOAT = 1e9

@tf.custom_gradient
def custom_norm(x):
    """Calculate the norm of a set of vector-like quantities, with some
    numeric stabilization applied to the gradient."""
    y = tf.linalg.norm(x, axis=-1, keepdims=True)

    def grad(dy):
        return dy * (x / (y + 1e-19))

    return y, grad

def bivec_dual(b):
    """scalar + bivector -> vector + trivector

    Calculates the dual of an input value, expressed as (scalar,
    bivector) with basis (1, e12, e13, e23).

    """
    swizzle = tf.constant([
        [0, 0, 0, -1],
        [0, 0, 1, 0],
        [0, -1, 0, 0],
        [1, 0, 0, 0]
    ], dtype=b.dtype)
    return tf.tensordot(b, swizzle, 1)

def vecvec(a, b):
    """vector*vector -> scalar + bivector

    Calculates the product of two vector inputs with basis (e1, e2,
    e3). Produces a (scalar, bivector) output with basis (1, e12, e13,
    e23).

    """
    products = a[..., tf.newaxis]*b[..., tf.newaxis, :]
    old_shape = tf.shape(products)
    new_shape = tf.concat([old_shape[:-2], [9]], -1)
    products = tf.reshape(products, new_shape)
    # 0 1 2
    # 3 4 5
    # 6 7 8
    swizzle = tf.constant([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, -1, 0, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 0, -1, 0],
        [0, 0, 0, -1],
        [1, 0, 0, 0],
    ], dtype=products.dtype)
    return tf.tensordot(products, swizzle, 1)

def vecvec_invariants(p):
    """Calculates rotation-invariant attributes of a (scalar, bivector) quantity.

    Returns a 2D output: the scalar and norm of the bivector.

    """
    result = [p[..., :1], custom_norm(p[..., 1:4])]
    return tf.concat(result, axis=-1)

def vecvec_covariants(p):
    """Calculates rotation-covariant attributes of a (scalar, bivector) quantity.

    Converts the bivector to a vector by taking the dual.

    """
    dual = bivec_dual(p)
    return dual[..., :3]

def bivecvec(p, c):
    """(scalar + bivector)*vector -> vector + trivector

    Calculates the product of a (scalar + bivector) and a vector. The
    two inputs are expressed in terms of the basis (1, e12, e13, e23)
    and (e1, e2, e3); the output is expressed in terms of the basis
    (e1, e2, e3, e123).

    """
    products = p[..., tf.newaxis]*c[..., tf.newaxis, :]
    old_shape = tf.shape(products)
    new_shape = tf.concat([old_shape[:-2], [12]], -1)
    products = tf.reshape(products, new_shape)
    # 0 1 2
    # 3 4 5
    # 6 7 8
    # 9 10 11
    swizzle = tf.constant([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, -1, 0, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 0, -1, 0],
        [0, 0, 0, -1],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 0, -1, 0],
        [0, 1, 0, 0],
    ], dtype=products.dtype)
    return tf.tensordot(products, swizzle, 1)

def bivecvec_invariants(q):
    """Calculates rotation-invariant attributes of a (vector, trivector) quantity.

    Returns a 2D output: the norm of the vector and the trivector.

    """
    result = [custom_norm(q[..., :3]), q[..., 3:4]]
    return tf.concat(result, axis=-1)

def bivecvec_covariants(q):
    """Calculates rotation-covariant attributes of a (vector, trivector) quantity.

    Returns the vector.

    """
    return q[..., :3]

def trivecvec(q, d):
    """(vector + trivector)*vector -> scalar + bivector

    Calculates the product of a (vector + trivector) and a vector. The
    two inputs are expressed in terms of the basis (e1, e2, e3, e123)
    and (e1, e2, e3); the output is expressed in terms of the basis
    (1, e12, e13, e23).

    """
    products = q[..., tf.newaxis]*d[..., tf.newaxis, :]
    old_shape = tf.shape(products)
    new_shape = tf.concat([old_shape[:-2], [12]], -1)
    products = tf.reshape(products, new_shape)
    # 0 1 2
    # 3 4 5
    # 6 7 8
    # 9 10 11
    swizzle = tf.constant([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, -1, 0, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 0, -1, 0],
        [0, 0, 0, -1],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 0, -1, 0],
        [0, 1, 0, 0],
    ], dtype=products.dtype)
    return tf.tensordot(products, swizzle, 1)

trivecvec_invariants = vecvec_invariants

trivecvec_covariants = vecvec_covariants

class GradientLayer(keras.layers.Layer):
    """Calculates the gradient of one input with respect to the other."""
    def call(self, inputs):
        return tf.gradients(inputs[0], inputs[1])

class MomentumNormalization(keras.layers.Layer):
    """Exponential decay normalization."""
    def __init__(self, momentum=.99, epsilon=1e-7, *args, **kwargs):
        self.momentum = momentum
        self.epsilon = epsilon
        super().__init__(*args, **kwargs)

    def build(self, input_shape):
        shape = [1]*len(input_shape)

        self.mu = self.add_weight(
            name='mu', shape=shape, initializer='zeros', trainable=False)
        self.sigma = self.add_weight(
            name='sigma', shape=shape, initializer='ones', trainable=False)

    def call(self, inputs, training=False):
        if training:
            mean = tf.math.reduce_mean(inputs, keepdims=True)
            std = tf.math.reduce_std(inputs, keepdims=True)
            self.mu.assign(self.momentum*self.mu + (1 - self.momentum)*mean)
            self.sigma.assign(self.momentum*self.sigma + (1 - self.momentum)*std)

        return (inputs - self.mu)/(self.sigma + self.epsilon)

    def get_config(self):
        result = super().get_config()
        result['momentum'] = self.momentum
        result['epsilon'] = self.epsilon
        return result

class NeighborhoodReduction(keras.layers.Layer):
    """Reduce values over the local neighborhood axis."""
    def __init__(self, mode='sum', *args, **kwargs):
        self.mode = mode

        super().__init__(*args, **kwargs)

    def call(self, inputs, mask=None):
        result = inputs
        if mask is not None:
            mask = tf.expand_dims(mask, -1)
            result = tf.where(mask, inputs, tf.zeros_like(inputs))

        if self.mode == 'sum':
            return tf.math.reduce_sum(result, axis=-2)
        elif self.mode == 'soft_max':
            return tf.math.reduce_logsumexp(result, axis=-2)
        else:
            raise NotImplementedError()

    def get_config(self):
        result = super().get_config()
        result['mode'] = self.mode
        return result

class PairwiseValueNormalization(keras.layers.Layer):
    """Normalize values over the -2 and -3 dimensions of output."""
    def call(self, inputs, training=False):
        mu = tf.math.reduce_mean(inputs, axis=(-2, -3), keepdims=True)
        sigma = tf.math.reduce_std(inputs, axis=(-2, -3), keepdims=True)
        return (inputs - mu)/sigma

class PairwiseVectorDifference(keras.layers.Layer):
    """Calculate the difference of all pairs of vectors in the neighborhood axis."""
    def call(self, inputs):
        return inputs[..., None, :] - inputs[..., None, :, :]

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return mask

        # (..., N, 3) -> (..., N)
        mask = tf.reduce_all(tf.not_equal(inputs, 0), axis=-1)
        # (..., N, N)
        mask = tf.logical_and(mask[..., None], mask[..., None, :])
        return mask

class PairwiseVectorDifferenceSum(keras.layers.Layer):
    """Calculate the symmetric difference and sum of all pairs of vectors in the neighborhood axis."""
    def call(self, inputs):
        return tf.concat([
            inputs[..., None, :] - inputs[..., None, :, :],
            inputs[..., None, :] + inputs[..., None, :, :]
        ], axis=-1)

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return mask

        mask = tf.reduce_any(tf.not_equal(inputs, 0), axis=-1)
        mask = tf.logical_and(mask[..., None], mask[..., None, :])
        return mask

_AttentionInputType = collections.namedtuple(
    'AttentionInputType', ['positions', 'values', 'weights'])

class VectorAttention(keras.layers.Layer):
    """Calculates geometric product attention.

    This layer implements a set of geometric products over all tuples
    of length `rank`, then sums over them using an attention mechanism
    to perform a permutation-covariant (`reduce=False`) or
    permutation-invariant (`reduce=True`) result.

    :param score_net: keras `Sequential` network producing logits for the attention mechanism
    :param value_net: keras `Sequential` network producing values in the embedding dimension of the network
    :param reduce: if `True`, produce a permutation-invariant result; otherwise, produce a permutation-covariant result
    :param merge_fun: Function used to merge the input values of each tuple before being passed to `join_fun`: 'mean' (no parameters) or 'concat' (learned projection for each tuple position)
    :param join_fun: Function used to join the representations of the rotation-invariant quantities (produced by `value_net`) and the tuple summary (produced by `merge_fun`): 'mean' (no parameters) or 'concat' (learned projection for each representation)
    :param rank: Degree of correlations to consider. 2 for pairwise attention, 3 for triplet-wise attention, and so on. Memory and computational complexity scales as `N**rank`

    """
    def __init__(self, score_net, value_net, reduce=True,
                 merge_fun='mean', join_fun='mean', rank=2,
                 *args, **kwargs):
        self.score_net = score_net
        self.value_net = value_net
        self.reduce = reduce
        self.merge_fun = merge_fun
        self.join_fun = join_fun
        self.rank = rank

        if merge_fun == 'mean':
            self.merge_fun_ = lambda *args: sum(args)/float(len(args))
        elif merge_fun == 'concat':
            self.merge_fun_ = lambda *args: sum(
                [tf.tensordot(x, b, 1) for (x, b) in zip(args, self.merge_kernels)])
        else:
            raise NotImplementedError()

        if join_fun == 'mean':
            self.join_fun_ = lambda *args: sum(args)/float(len(args))
        elif join_fun == 'concat':
            self.join_fun_ = lambda *args: sum(
                [tf.tensordot(x, b, 1) for (x, b) in zip(args, self.join_kernels)])
        else:
            raise NotImplementedError()

        super().__init__(*args, **kwargs)

    def build(self, input_shape):
        shapes = self._parse_inputs(input_shape)
        r_shape = shapes.positions
        v_shape = shapes.values

        self.merge_kernels = None
        if self.merge_fun == 'concat':
            stdev = tf.sqrt(2./self.rank/v_shape[-1])
            self.merge_kernels = [self.add_weight(
                name='merge_kernel_{}'.format(i), shape=(v_shape[-1], v_shape[-1]),
                initializer=keras.initializers.RandomNormal(stddev=stdev)
            ) for i in range(self.rank)]

        self.join_kernels = None
        if self.join_fun == 'concat':
            # always joining neighborhood values and invariant values
            stdev = tf.sqrt(2./2/v_shape[-1])
            self.join_kernels = [self.add_weight(
                name='join_kernel_{}'.format(i), shape=(v_shape[-1], v_shape[-1]),
                initializer=keras.initializers.RandomNormal(stddev=stdev)
            ) for i in range(2)]

        return super().build(input_shape)

    def _expand_products(self, rs, vs):
        broadcast_indices = []
        for i in range(1, self.rank + 1):
            index = [Ellipsis] + [None]*(self.rank) + [slice(None)]
            index[-i - 1] = slice(None)
            broadcast_indices.append(index)

        expanded_vs = [vs[index] for index in broadcast_indices]
        expanded_rs = [rs[index] for index in broadcast_indices]

        product_funs = itertools.chain(
            [vecvec], itertools.cycle([bivecvec, trivecvec]))
        invar_funs = itertools.chain(
            [vecvec_invariants], itertools.cycle([bivecvec_invariants, trivecvec_invariants]))
        covar_funs = itertools.chain(
            [vecvec_covariants], itertools.cycle([bivecvec_covariants, trivecvec_covariants]))

        left = expanded_rs[0]

        invar_fn = custom_norm
        covar_fn = lambda x: x
        for (product_fn, invar_fn, covar_fn, right) in zip(
                product_funs, invar_funs, covar_funs, expanded_rs[1:]):
            left = product_fn(left, right)

        invar = invar_fn(left)
        covar = covar_fn(left)

        return broadcast_indices, invar, covar, expanded_vs

    def _intermediates(self, inputs, mask=None):
        parsed_inputs = self._parse_inputs(inputs)
        positions = parsed_inputs.positions
        values = parsed_inputs.values
        (broadcast_indices, invariants, _, expanded_values) = \
            self._expand_products(positions, values)
        neighborhood_values = self.merge_fun_(*expanded_values)
        invar_values = self.value_net(invariants)

        joined_values = self.join_fun_(invar_values, neighborhood_values)
        new_values = joined_values

        tuple_weights = self._make_tuple_weights(broadcast_indices, parsed_inputs.weights)
        scores = self.score_net(joined_values)
        old_shape = tf.shape(scores)

        if mask is not None:
            parsed_mask = self._parse_inputs(mask)
            position_mask = parsed_mask.positions
            value_mask = parsed_mask.values
            if position_mask is not None:
                position_mask = tf.expand_dims(position_mask, -1)
                position_mask = tf.reduce_all([position_mask[idx] for idx in broadcast_indices[:-1]], axis=0)
            else:
                position_mask = True
            if value_mask is not None:
                value_mask = tf.expand_dims(value_mask, -1)
                value_mask = tf.reduce_all([value_mask[idx] for idx in broadcast_indices[:-1]], axis=0)
            else:
                value_mask = True
            product_mask = tf.logical_and(position_mask, value_mask)
            scores = tf.where(product_mask, scores, -HUGE_FLOAT)

        if self.reduce:
            dims = -(self.rank + 1)
            reduce_axes = tuple(-i - 2 for i in range(self.rank))
        else:
            dims = -self.rank
            reduce_axes = tuple(-i - 2 for i in range(self.rank - 1))

        shape = tf.concat([old_shape[:dims],
                           tf.math.reduce_prod(old_shape[dims:], keepdims=True)], -1)
        scores = tf.reshape(scores, shape)
        attention = tf.reshape(tf.nn.softmax(scores), old_shape)
        output = tf.reduce_sum(attention*tuple_weights*new_values, reduce_axes)

        return dict(attention=attention, output=output, invariants=invar_values)

    def _make_tuple_weights(self, broadcast_indices, weights):
        if isinstance(weights, int):
            return weights
        expanded_weights = [weights[..., None][idx] for idx in broadcast_indices]
        result = 1
        for w in expanded_weights:
            result = result*w
        return tf.math.pow(result, 1./self.rank)

    def _parse_inputs(self, inputs):
        if len(inputs) == 2:
            (r, v) = inputs
            w = 1
        elif len(inputs) == 3:
            (r, v, w) = inputs
        else:
            raise NotImplementedError(inputs)
        return _AttentionInputType(r, v, w)

    def call(self, inputs, return_invariants=False, return_attention=False, mask=None):
        intermediates = self._intermediates(inputs, mask)
        result = [intermediates['output']]
        if return_invariants:
            result.append(intermediates['invariants'])
        if return_attention:
            result.append(intermediates['attention'])

        if len(result) > 1:
            return tuple(result)
        else:
            return result[0]

    def compute_mask(self, inputs, mask=None):
        if not self.reduce or mask is None:
            return mask

        parsed_mask = self._parse_inputs(mask)
        position_mask = parsed_mask.positions
        value_mask = parsed_mask.values
        if position_mask is not None:
            return tf.reduce_any(position_mask, axis=-1)
        else:
            return tf.reduce_any(value_mask, axis=-1)

    @classmethod
    def from_config(cls, config):
        new_config = dict(config)
        for key in ('score_net', 'value_net'):
            new_config[key] = keras.models.Sequential.from_config(new_config[key])
        return cls(**new_config)

    def get_config(self):
        result = super().get_config()
        result['score_net'] = self.score_net.get_config()
        result['value_net'] = self.value_net.get_config()
        result['reduce'] = self.reduce
        result['merge_fun'] = self.merge_fun
        result['join_fun'] = self.join_fun
        result['rank'] = self.rank
        return result

class Vector2VectorAttention(VectorAttention):
    """Adaptation of `VectorAttention` to produce vector outputs.

    Cf. `VectorAttention` for most arguments.

    :param scale_net: keras `Sequential` network producing scalar rescaling factors for each vector
    :param use_product_vectors: Use the vector extracted from the geometric product of each tuple within the attention mechanism
    :param use_input_vectors: Use the input vectors for each tuple within the attention mechanism. A scalar weight will be learned for each (i, j, k, ...) element of the tuple
    :param learn_vector_projection: Use a learned linear projection combination factor for the summary value (for the product vector) and the input values (for the input vectors)

    """

    def __init__(self, score_net, value_net, scale_net, *args,
                 use_product_vectors=True, use_input_vectors=False,
                 learn_vector_projection=False,
                 **kwargs):
        super().__init__(score_net, value_net, *args, **kwargs)

        self.scale_net = scale_net
        self.use_product_vectors = use_product_vectors
        self.use_input_vectors = use_input_vectors
        self.learn_vector_projection = learn_vector_projection

        self.input_vector_count = (
            int(self.use_product_vectors) + self.rank*self.use_input_vectors)
        if self.input_vector_count < 1:
            raise ValueError('At least one of use_product_vectors or '
                             'use_input_vectors must be True')

    def _covariants(self, covariants_, positions, broadcast_indices, expanded_values,
                    joined_values):
        covariant_vectors = []
        input_values = []
        if self.use_product_vectors:
            covariant_vectors.append(covariants_)
            input_values.append(joined_values)
        if self.use_input_vectors:
            covariant_vectors.extend(
                [positions[idx] for idx in broadcast_indices])
            input_values.extend(expanded_values)

        scalars = self.vector_kernels
        if self.learn_vector_projection:
            scalars = [tf.tensordot(v, kernel, 1) + self.vector_biases[i]
                       for i, (v, kernel) in
                       enumerate(zip(input_values, self.vector_kernels))]

        covariants = [
            vec*scalars[i] for (i, vec) in enumerate(covariant_vectors)]
        return sum(covariants)

    def _intermediates(self, inputs, mask=None):
        parsed_inputs = self._parse_inputs(inputs)
        positions = parsed_inputs.positions
        values = parsed_inputs.values
        (broadcast_indices, invariants, covariants, expanded_values) = \
            self._expand_products(positions, values)
        neighborhood_values = self.merge_fun_(*expanded_values)
        invar_values = self.value_net(invariants)

        joined_values = self.join_fun_(invar_values, neighborhood_values)
        covariants = self._covariants(
            covariants, positions, broadcast_indices, expanded_values, joined_values)

        tuple_weights = self._make_tuple_weights(broadcast_indices, parsed_inputs.weights)
        scales = self.scale_net(joined_values)
        scores = self.score_net(joined_values)
        old_shape = tf.shape(scores)

        if mask is not None:
            parsed_mask = self._parse_inputs(mask)
            position_mask = parsed_mask.positions
            value_mask = parsed_mask.values
            if position_mask is not None:
                position_mask = tf.expand_dims(position_mask, -1)
                position_mask = tf.reduce_all([position_mask[idx] for idx in broadcast_indices[:-1]], axis=0)
            else:
                position_mask = True
            if value_mask is not None:
                value_mask = tf.expand_dims(value_mask, -1)
                value_mask = tf.reduce_all([value_mask[idx] for idx in broadcast_indices[:-1]], axis=0)
            else:
                value_mask = True
            product_mask = tf.logical_and(position_mask, value_mask)
            scores = tf.where(product_mask, scores, -HUGE_FLOAT)

        if self.reduce:
            dims = -(self.rank + 1)
            reduce_axes = tuple(-i - 2 for i in range(self.rank))
        else:
            dims = -self.rank
            reduce_axes = tuple(-i - 2 for i in range(self.rank - 1))

        shape = tf.concat([old_shape[:dims], tf.math.reduce_prod(old_shape[dims:], keepdims=True)], -1)
        scores = tf.reshape(scores, shape)
        attention = tf.reshape(tf.nn.softmax(scores), old_shape)
        output = tf.reduce_sum(attention*tuple_weights*covariants*scales, reduce_axes)

        return dict(attention=attention, output=output, invariants=invariants)

    def build(self, input_shape):
        parsed_inputs = self._parse_inputs(input_shape)
        value_shape = parsed_inputs.values
        value_dim = value_shape[-1]

        self.vector_kernels = [1]
        self.vector_biases = [0]
        if self.use_input_vectors:
            if self.learn_vector_projection:
                self.vector_kernels = [self.add_weight(
                    name='vector_kernels_{}'.format(i), shape=[value_dim, 1])
                    for i in range(self.input_vector_count)]
                self.vector_biases = self.add_weight(
                    name='vector_biases', shape=[self.input_vector_count],
                    initializer=keras.initializers.RandomNormal(mean=0.))
            else:
                self.vector_kernels = self.add_weight(
                    name='vector_kernels', shape=[self.input_vector_count],
                    initializer=keras.initializers.RandomNormal(mean=0.))
        return super().build(input_shape)

    @classmethod
    def from_config(cls, config):
        new_config = dict(config)
        for key in ('scale_net',):
            new_config[key] = keras.models.Sequential.from_config(new_config[key])
        return super(Vector2VectorAttention, cls).from_config(new_config)

    def get_config(self):
        result = super().get_config()
        result['scale_net'] = self.scale_net.get_config()
        result['use_product_vectors'] = self.use_product_vectors
        result['use_input_vectors'] = self.use_input_vectors
        result['learn_vector_projection'] = self.learn_vector_projection
        return result



from tensorflow import keras

class ScaledMSE(keras.metrics.MeanSquaredError):
    """Calculate the mean squared error with a specified scaling factor applied."""
    def __init__(self, scale=1., *args, **kwargs):
        self.scale = scale
        super().__init__(*args, **kwargs)

    def result(self):
        return super().result()*self.scale**2

    def get_config(self):
        result = super().get_config()
        result['scale'] = self.scale
        return result

class ScaledMAE(keras.metrics.MeanAbsoluteError):
    """Calculate the mean absolute error with a specific scaling factor applied."""
    def __init__(self, scale=1., *args, **kwargs):
        self.scale = scale
        super().__init__(*args, **kwargs)

    def result(self):
        return super().result()*self.scale

    def get_config(self):
        result = super().get_config()
        result['scale'] = self.scale
        return result



import os
import subprocess

import flowws
from flowws import Argument as Arg
import numpy as np
from tensorflow import keras

from .internal import ScaledMSE, ScaledMAE

@flowws.add_stage_arguments
class MD17(flowws.Stage):
    """Load data from the MD17 dataset for molecular force regression.

    This module downloads files from the MD17 dataset as required. It
    randomly selects molecule snapshots for the specified molecules to
    produce training, testing, and validation sets.

    For benzene in particular, the "old" benzene calculation of the
    original MD17 dataset is used.

    """

    ARGS = [
        Arg('n_train', '-n', int, 1000,
            help='Number of frames to take for training'),
        Arg('n_val', None, int, 1000,
            help='Number of frames to take for validation'),
        Arg('n_test', None, int, 1000,
            help='Number of frames to take for testing'),
        Arg('cache_dir', '-c', str, '/tmp/md17',
            help='Directory to store trajectory data'),
        Arg('molecules', '-m', [str], [],
            help='List of molecules to use'),
        Arg('seed', '-s', int, 13,
            help='Random number seed to use'),
        Arg('units', None, str, 'meV',
            help='Energy units to use (meV or kcal/mol)'),
        Arg('y_scale_reduction', None, float, 16,
            help='Factor by which to scale forces for training purposes'),
        Arg('x_scale_reduction', None, float,
            help='Factor by which to scale input distances for training purposes'
            ' (negative to auto-scale from training data)')
    ]

    def run(self, scope, storage):
        np.random.seed(self.arguments['seed'])

        energy_conversion = 1.
        if self.arguments['units'] == 'meV':
            energy_conversion = 43.36
        elif self.arguments['units'] == 'kcal/mol':
            pass
        else:
            raise NotImplementedError(self.arguments['units'])

        loaded_files = {}
        train_indices, val_indices, test_indices = {}, {}, {}
        N_train = self.arguments['n_train']
        N_val = self.arguments['n_val']
        N_test = self.arguments['n_test']

        max_atoms = 0
        seen_types = set()

        for name in self.arguments['molecules']:
            fname = self._download(name)
            loaded_files[fname] = data = np.load(fname, mmap_mode='r')
            (frames, size, types) = self.get_trajectory_size_types(data)
            max_atoms = max(max_atoms, size)

            indices = np.arange(frames)
            np.random.shuffle(indices)
            train_indices[fname] = indices[:N_train]
            val_indices[fname] = indices[N_train:N_train + N_val]
            test_indices[fname] = indices[N_train + N_val:N_train + N_val + N_test]
            seen_types.update(types)

        all_types = [0] + list(sorted(seen_types))
        type_map = {t: i for (i, t) in enumerate(all_types)}
        num_types = len(all_types)

        datasets = {}
        for name in ['train', 'val', 'test']:
            dset_xs, dset_ts, dset_ys = [], [], []
            for fname in sorted(loaded_files):
                indices = locals()['{}_indices'.format(name)]
                (xs, ts), Us = self.get_encoding(
                    loaded_files[fname], max_atoms, type_map, indices[fname], energy_conversion)
                dset_xs.append(xs)
                dset_ts.append(ts)
                dset_ys.append(Us)

            dset_xs = np.concatenate(dset_xs, axis=0)
            dset_ts = np.concatenate(dset_ts, axis=0)
            dset_ys = np.concatenate(dset_ys, axis=0)

            indices = np.arange(len(dset_xs))
            np.random.shuffle(indices)
            dset_xs = dset_xs[indices]
            dset_ts = dset_ts[indices]
            dset_ys = dset_ys[indices]

            datasets[name] = (dset_xs, dset_ts, dset_ys)

        yscale = np.std(datasets['train'][-1])*self.arguments['y_scale_reduction']

        for (_, _, y) in datasets.values():
            y /= yscale

        scaled_mse = ScaledMSE(yscale)
        scaled_mae = ScaledMAE(yscale)

        xscale = 1.
        if 'x_scale_reduction' in self.arguments:
            xscale = self.arguments['x_scale_reduction']

            if xscale <= 0:
                delta = datasets['train'][0] - datasets['train'][0][:, :1]
                delta = delta.reshape((-1, 3))
                filt = np.logical_and(
                    np.any(datasets['train'][0].reshape((-1, 3)) != 0, axis=-1),
                    np.any(delta != 0, axis=-1))
                delta = np.linalg.norm(delta[filt], axis=-1)
                xscale = np.std(delta)

            for (x, _, _) in datasets.values():
                x /= xscale

        scope['y_scale'] = yscale
        scope['x_scale'] = xscale
        scope['neighborhood_size'] = max_atoms
        scope['num_types'] = num_types
        scope['x_train'] = datasets['train'][:2]
        scope['y_train'] = datasets['train'][-1]
        scope['x_test'] = datasets['test'][:2]
        scope['y_test'] = datasets['test'][-1]
        scope['validation_data'] = (datasets['val'][:2], datasets['val'][-1])
        scope['type_map'] = type_map
        scope.setdefault('metrics', []).extend([scaled_mse, scaled_mae])

    def _download(self, name):
        url = self.get_url(name)
        fname = url.split('/')[-1]

        if fname not in os.listdir(self.arguments['cache_dir']):
            command = ['wget', '-c', '-P', self.arguments['cache_dir'], url]
            subprocess.check_call(command)

        return os.path.join(self.arguments['cache_dir'], fname)

    @staticmethod
    def get_encoding(data, max_atoms, type_map, indices=None, energy_conversion=1.):
        coords = data['R']
        forces = data['F']*energy_conversion
        types = np.zeros(max_atoms, dtype=np.uint32)
        types[:coords.shape[1]] = [type_map[t] for t in data['z']]

        if indices is not None:
            coords = coords[indices]
            forces = forces[indices]

        rs = np.zeros((len(coords), max_atoms, 3))
        rs[:, :coords.shape[1], :] = coords

        Fs = np.zeros((len(coords), max_atoms, 3))
        Fs[:, :coords.shape[1], :] = forces

        types_onehot = np.eye(len(type_map))[types]
        types_onehot = np.tile(types_onehot[np.newaxis, ...], (len(coords), 1, 1))
        return (rs, types_onehot), Fs

    @staticmethod
    def get_trajectory_size_types(data):
        (frames, size, _) = data['R'].shape
        return (frames, size, data['z'])

    @staticmethod
    def get_url(name):
        remap = dict(benzene='benzene_old_dft.npz')

        base = 'http://quantum-machine.org/gdml/data/npz/'

        return base + remap.get(name, '{}_dft.npz'.format(name.split('_')[0]))



from flowws import try_to_import

from .MD17 import MD17
from .PDBCache import PDBCache
PDBCoarseGrained = try_to_import('.PDBCoarseGrained', 'PDBCoarseGrained', __name__)
PyriodicDataset = try_to_import('.PyriodicDataset', 'PyriodicDataset', __name__)
from .RMD17 import RMD17




import flowws
from flowws import Argument as Arg
import numpy as np
from tensorflow import keras

def convert_generator(gen):
    for (r, v, cv), y in gen:
        rv = np.concatenate([r, v], axis=-1)
        yield (rv, cv), y

@flowws.add_stage_arguments
class PDBInverseCoarseGrainTransformer(flowws.Stage):
    """Build a transformer-based (naive, non rotation-equivariant) neural network for a coarse-grain backmapping task.

    This module specifies the architecture of a network to produce
    atomic coordinates from a set of coarse-grained beads.

    """

    ARGS = [
        Arg('n_dim', '-n', int, 32,
            help='Working dimensionality of point representations'),
        Arg('dilation', None, float, 2,
            help='Working dimension dilation factor for MLP components'),
        Arg('n_blocks_coarse', None, int, 2,
            help='Number of deep blocks to use in the coarse-grain space'),
        Arg('n_blocks_fine', None, int, 2,
            help='Number of deep blocks to use in the coarse-grain space'),
        Arg('block_nonlinearity', None, bool, True,
            help='If True, add a nonlinearity to the end of each block'),
        Arg('residual', '-r', bool, True,
            help='If True, use residual connections within blocks'),
        Arg('activation', '-a', str, 'swish',
            help='Activation function to use inside the network'),
    ]

    def run(self, scope, storage):
        n_dim = self.arguments['n_dim']

        for name in ['train', 'validation', 'test']:
            key = '{}_generator'.format(name)
            if key in scope:
                scope[key] = convert_generator(scope[key])

        train_data = scope['train_generator']
        sample_batch = next(train_data)

        rv_in = keras.layers.Input(sample_batch[0][0].shape[1:], name='rv')
        cv_in = keras.layers.Input(sample_batch[0][1].shape[1:], name='child_t')

        cv_emb = keras.layers.Embedding(len(scope['child_type_names']), n_dim, mask_zero=True)(cv_in)

        dilation_dim = round(n_dim*self.arguments['dilation'])

        def make_block(right, last=None, residual=False):
            if last is None:
                last = right

            residual_in = last
            last = keras.layers.Attention()([right, last, last])

            if self.arguments['block_nonlinearity']:
                last = keras.layers.Dense(dilation_dim)(last)
                last = keras.layers.Activation(self.arguments['activation'])(last)
                last = keras.layers.Dense(n_dim)(last)

            if residual:
                last = last + residual_in

            return last

        last = keras.layers.Dense(n_dim)(rv_in)
        for _ in range(self.arguments['n_blocks_coarse']):
            last = make_block(last, residual=self.arguments['residual'])

        last = make_block(cv_emb, last, False)

        for _ in range(self.arguments['n_blocks_fine']):
            last = make_block(last, residual=self.arguments['residual'])

        last = keras.layers.Dense(dilation_dim)(last)
        last = keras.layers.Activation(self.arguments['activation'])(last)
        vec = keras.layers.Dense(3)(last)

        scope['input_symbol'] = [rv_in, cv_in]
        scope['output'] = vec
        scope['loss'] = 'mse'



import collections

import flowws
from flowws import Argument as Arg
import freud
import numpy as np
import pyriodic

def encode_types(source_types, dest_types, N, max_types):
    onehot_src = np.eye(max_types)[source_types]
    onehot_dest = np.eye(max_types)[dest_types]

    minus = onehot_dest - onehot_src
    minus = minus.reshape((-1, N, max_types))
    plus = onehot_dest + onehot_src
    plus = plus.reshape((-1, N, max_types))

    return np.concatenate([minus, plus], axis=-1)

@flowws.add_stage_arguments
class PyriodicDataset(flowws.Stage):
    """Load crystal structures from `pyriodic` for training data.

    This module takes a specified set of structures from the
    `pyriodic` default database, replicates them up to the given size,
    applies Gaussian noise to each structure one or more times, and
    extracts local environments of these structures for training data.

    """

    ARGS = [
        Arg('num_neighbors', '-n', int, 12,
           help='Number of nearest neighbors to use'),
        Arg('structures', '-s', [str],
           help='Name of structures to take'),
        Arg('size', None, int, 512,
           help='Number of particles to replicate structures up to'),
        Arg('noise', None, [float], [1e-2, 5e-2, .1],
           help='Noise standard deviation to apply to structures'),
        Arg('test_fraction', '-t', float, 0,
           help='Fraction of data to hold back as test data'),
        Arg('seed', None, int, 13,
           help='Random seed to use for shuffling training data'),
    ]

    def run(self, scope, storage):
        np.random.seed(self.arguments['seed'])

        xs = []
        ts = []
        ys = []

        name_map = collections.defaultdict(lambda: len(name_map))

        structures = list(self.arguments['structures'])
        max_types = 0
        for name in structures:
            for (structure,) in pyriodic.db.query('select structure from unit_cells where name = ?', (name,)):
                pass
            max_types = max(max_types, len(set(structure.types)))

        for name in structures:
            for (structure,) in pyriodic.db.query('select structure from unit_cells where name = ?', (name,)):
                pass

            if name in name_map:
                continue

            for noise in self.arguments['noise']:
                structure = structure.rescale_shortest_distance(1)
                structure = structure.replicate_upto(self.arguments['size'])
                structure = structure.add_gaussian_noise(noise)

                q = freud.locality.AABBQuery(structure.box, structure.positions)
                qr = q.query(
                    structure.positions, dict(
                        num_neighbors=self.arguments['num_neighbors'], exclude_ii=True))
                nl = qr.toNeighborList()
                rijs = (structure.positions[nl.point_indices] -
                        structure.positions[nl.query_point_indices])
                fbox = freud.box.Box(*structure.box)
                rijs = fbox.wrap(rijs)
                rijs = rijs.reshape(
                    (len(structure), self.arguments['num_neighbors'], 3))

                tijs = encode_types(
                    structure.types[nl.query_point_indices],
                    structure.types[nl.point_indices],
                    self.arguments['num_neighbors'], max_types)

                shuf = np.arange(len(rijs))
                np.random.shuffle(shuf)
                shuf = shuf[:self.arguments['size']]

                xs.append(rijs[shuf])
                ts.append(tijs[shuf])
                ys.append(name_map[name])

        ys = np.repeat(ys, [len(v) for v in xs])
        xs = np.concatenate(xs, axis=0)
        ts = np.concatenate(ts, axis=0)

        shuf = np.arange(len(xs))
        np.random.shuffle(shuf)
        N_test = int(self.arguments['test_fraction']*len(shuf))
        test_split, train_split = shuf[:N_test], shuf[N_test:]
        xs_test = xs[test_split]
        ts_test = ts[test_split]
        ys_test = ys[test_split]
        xs = xs[train_split]
        ts = ts[train_split]
        ys = ys[train_split]

        scope['x_train'] = (xs, ts)
        scope['y_train'] = ys
        scope['x_test'] = (xs_test, ts_test)
        scope['y_test'] = ys_test
        scope['num_classes'] = len(name_map)
        scope['type_map'] = dict(name_map)
        scope['neighborhood_size'] = self.arguments['num_neighbors']



import flowws
from flowws import Argument as Arg
import numpy as np
from tensorflow import keras

from .internal import VectorAttention

@flowws.add_stage_arguments
class CrystalStructureClassification(flowws.Stage):
    """Build a geometric attention network for the structure identification task.

    This module specifies the architecture of a network to classify
    local environments of crystal structures in a rotation-invariant
    manner.

    """

    ARGS = [
        Arg('rank', None, int, 2,
            help='Degree of correlations (n-vectors) to consider'),
        Arg('n_dim', '-n', int, 32,
            help='Working dimensionality of point representations'),
        Arg('dilation', None, float, 2,
            help='Working dimension dilation factor for MLP components'),
        Arg('merge_fun', '-m', str, 'mean',
            help='Method to merge point representations'),
        Arg('join_fun', '-j', str, 'mean',
            help='Method to join invariant and point representations'),
        Arg('dropout', '-d', float, 0,
            help='Dropout rate to use, if any'),
        Arg('n_blocks', '-b', int, 2,
            help='Number of deep blocks to use'),
        Arg('block_nonlinearity', None, bool, True,
            help='If True, add a nonlinearity to the end of each block'),
        Arg('residual', '-r', bool, True,
            help='If True, use residual connections within blocks'),
        Arg('activation', '-a', str, 'relu',
            help='Activation function to use inside the network'),
        Arg('final_activation', None, str, 'relu',
            help='Final activation function to use within the network'),
    ]

    def run(self, scope, storage):
        n_dim = self.arguments['n_dim']
        dilation_dim = int(np.round(n_dim*self.arguments['dilation']))

        def make_scorefun():
            layers = [
                keras.layers.Dense(dilation_dim),
                keras.layers.Activation(self.arguments['activation'])
            ]

            if self.arguments.get('dropout', 0):
                layers.append(keras.layers.Dropout(self.arguments['dropout']))

            layers.append(keras.layers.Dense(1))
            return keras.models.Sequential(layers)

        def make_valuefun():
            layers = [
                keras.layers.Dense(dilation_dim),
                keras.layers.LayerNormalization(),
                keras.layers.Activation(self.arguments['activation']),
            ]

            if self.arguments.get('dropout', 0):
                layers.append(keras.layers.Dropout(self.arguments['dropout']))

            layers.append(keras.layers.Dense(n_dim))
            return keras.models.Sequential(layers)

        def make_block(last):
            residual_in = last
            last = VectorAttention(
                make_scorefun(), make_valuefun(), False, rank=self.arguments['rank'],
                join_fun=self.arguments['join_fun'],
                merge_fun=self.arguments['merge_fun'])([x_in, last])

            if self.arguments['block_nonlinearity']:
                last = make_valuefun()(last)

            if self.arguments['residual']:
                last = last + residual_in

            return last

        (xs, ts) = scope['x_train']
        x_in = keras.layers.Input(xs[0].shape)
        v_in = keras.layers.Input(ts[0].shape)

        last = keras.layers.Dense(n_dim)(v_in)
        for _ in range(self.arguments['n_blocks']):
            last = make_block(last)

        (last, ivs, att) = VectorAttention(
            make_scorefun(), make_valuefun(), True, name='final_attention',
            rank=self.arguments['rank'], join_fun=self.arguments['join_fun'],
            merge_fun=self.arguments['merge_fun'])(
            [x_in, last], return_invariants=True, return_attention=True)
        last = keras.layers.Dense(dilation_dim, name='final_mlp')(last)
        if self.arguments.get('dropout', 0):
            last = keras.layers.Dropout(self.arguments['dropout'])(last)
        last = keras.layers.Activation(self.arguments['final_activation'])(last)
        last = keras.layers.Dense(scope['num_classes'], activation='softmax')(last)

        scope['input_symbol'] = [x_in, v_in]
        scope['output'] = last
        scope['loss'] = 'sparse_categorical_crossentropy'
        scope['attention_model'] = keras.models.Model([x_in, v_in], att)
        scope['invariant_model'] = keras.models.Model([x_in, v_in], ivs)
        scope.setdefault('metrics', []).append('accuracy')



import collections
import itertools
import functools
import json

import flowws
from flowws import Argument as Arg
import keras_gtar
import numpy as np

LoadResult = collections.namedtuple(
    'LoadResult', ['att_model', 'train_data', 'test_data', 'batch_size'])

@functools.lru_cache(maxsize=1)
def load_model(fname, path_substitutions):
    path_substitutions = dict(path_substitutions)

    with keras_gtar.Trajectory(fname, 'r') as traj:
        weights = traj.get_weights()
        workflow_json = traj.handle.readStr('workflow.json')

    w = flowws.Workflow.from_JSON(json.loads(workflow_json))
    stages = []
    for stage in w.stages:
        stage_json = stage.to_JSON()
        stages.append(stage)

        for (name, val) in stage_json['arguments'].items():
            if isinstance(val, str):
                stage.arguments[name] = path_substitutions.get(
                    stage.arguments[name], stage.arguments[name])

        if stage_json['type'] == 'Train':
            stage.arguments['epochs'] = 0
            break
    new_w = flowws.Workflow(stages)
    scope = new_w.run()
    model = scope['model']
    model.set_weights(weights)

    batch_size = new_w.stages[-1].arguments['batch_size']
    train_xs, train_ts = scope['x_train']
    attention_model = scope['attention_model']
    try:
        test_data = scope['x_test'], scope['y_test']
    except KeyError:
        test_data = None

    return LoadResult(
        attention_model, (train_xs, train_ts),
        test_data, batch_size)

@flowws.add_stage_arguments
class ParticleAttentionViewer(flowws.Stage):
    """Plot pairwise attention for small particle neighborhoods.

    This module creates a plato scene to visualize local particle
    environments and the attention map of a trained model. It loads a
    gtar file containing a trained model and flowws workflow
    definition (using `PyriodicDataset`, for example) and enables
    visualization of the coordinates and pairwise particle attention.

    Workflows run on different systems will likely require use of the
    `path_substitutions` argument for their data-loading modules.

    """

    ARGS = [
        Arg('filename', '-f', str,
            help='Saved model to open'),
        Arg('path_substitutions', '-p', [(str, str)],
            help='Paths to replace in stage descriptions'),
        Arg('frame', None, int, 0,
            help='Frame (data index) to visualize'),
        Arg('additive_rendering', '-a', bool, True,
            help='If True, use additive rendering'),
        Arg('fast_antialiasing', None, bool, False,
            help='Use Fast Approximate Antialiasing (FXAA)'),
        Arg('ambient_occlusion', None, bool, False,
            help='Use Screen Space Ambient Occlusion (SSAO)'),
        Arg('filter_value', None, float,
            help='Filter bonds by minimum attention value'),
        Arg('value_scale', None, float, 0.,
            help='Factor to modify bond values by (after filter_value; in log-space)'),
        Arg('diameter_scale', '-d', float, 1.,
            help='Size scale for atomic diameters'),
        Arg('clip_min', None, float, 0.,
            help='Minimum value of attention weights (after scaling by value_scale) to clip to'),
        Arg('clip_max', None, float, 1.,
            help='Maximum value of attention weights (after scaling by value_scale) to clip to'),
        Arg('cmap_s', None, float, 0., valid_values=flowws.Range(-3., 3., True),
            help='Bond colormap s argument (0, 1, 2) -> (blue, red, green)'),
        Arg('cmap_h', None, float, 1.2, valid_values=flowws.Range(0., 2., True),
            help='Bond colormap h argument controlling saturation'),
        Arg('cmap_r', None, float, 1., valid_values=flowws.Range(0., 8., True),
            help='Bond colormap r argument controlling number of rotations to make'),
    ]

    def run(self, scope, storage):
        loaded = load_model(
            self.arguments['filename'], tuple(self.arguments['path_substitutions']))
        self.attention_model = loaded.att_model
        self.train_data = loaded.train_data
        self.test_data = loaded.test_data
        self.batch_size = loaded.batch_size

        self.arg_specifications['frame'].valid_values = flowws.Range(
            0, len(self.train_data[0]), (True, False))

        scope.setdefault('visuals', []).append(self)

    def draw_plato(self):
        import plato, plato.draw as draw

        train_xs, train_ts = self.train_data
        i = self.arguments['frame']

        features = {}
        if self.arguments.get('additive_rendering', True):
            features['additive_rendering'] = True
        if self.arguments.get('fast_antialiasing', False):
            features['fxaa'] = True
        if self.arguments.get('ambient_occlusion', False):
            features['ssao'] = True

        prim = draw.Spheres()
        edges = draw.Lines(cap_mode=1)
        scene = draw.Scene([prim, edges], zoom=3.5, features=features)

        xs = train_xs[i].copy()

        attention = self.attention_model.predict(
            (train_xs[i:i+1], train_ts[i:i+1]))[0, ..., 0]

        cmap_kwargs = dict(
            s=self.arguments['cmap_s'], h=self.arguments['cmap_h'], r=self.arguments['cmap_r'])

        plusminus = train_ts[i]
        minus, plus = plusminus[:, :plusminus.shape[1]//2], plusminus[:, plusminus.shape[1]//2:]
        types = np.argmax(plus - minus, axis=-1)

        prim.positions = xs
        prim.diameters = 1.*self.arguments['diameter_scale']
        colors = plato.cmap.cubehelix(np.diag(attention).copy(), **cmap_kwargs)
        prim.colors = colors*np.exp(self.arguments['value_scale'])

        start_points = []
        end_points = []
        colors = []
        for (ii, jj) in itertools.product(range(len(xs)), range(len(xs))):
            if self.arguments.get('filter_value', 0):
                if attention[ii, jj] < self.arguments['filter_value']:
                    continue
            start_points.append(xs[ii])
            end_points.append(xs[jj])
            colors.append(attention[ii, jj])
        colors = np.array(colors)
        colors = np.clip(colors, self.arguments['clip_min'], self.arguments['clip_max'])
        colors = plato.cmap.cubehelix(colors, **cmap_kwargs)
        edges.start_points = start_points
        edges.end_points = end_points
        edges.widths = .125
        edges.colors = colors*np.exp(self.arguments['value_scale'])

        return scene



import collections
import functools
import json

import flowws
from flowws import Argument as Arg
import keras_gtar
import numpy as np

elt_colors = {
    'H': (.8, .8, .8, 1),
    'C': (.5, .5, .5, 1),
    'N': (.5, .7, .8, 1),
    'O': (.8, .5, .5, 1),
    'P': (.8, .7, 0, 1),
    'S': (.8, .8, 0, 1),
}

elt_radii = {
    'H': .25,
    'C': .7,
    'N': .65,
    'O': .6,
    'P': 1.,
    'S': 1.,
}

LoadResult = collections.namedtuple(
    'LoadResult', ['model', 'scope'])

@functools.lru_cache(maxsize=1)
def load_model(fname, path_substitutions):
    path_substitutions = dict(path_substitutions)

    with keras_gtar.Trajectory(fname, 'r') as traj:
        weights = traj.get_weights()
        workflow_json = traj.handle.readStr('workflow.json')

    w = flowws.Workflow.from_JSON(json.loads(workflow_json))
    stages = []
    for stage in w.stages:
        stage_json = stage.to_JSON()
        stages.append(stage)

        for (name, val) in stage_json['arguments'].items():
            if isinstance(val, str):
                stage.arguments[name] = path_substitutions.get(
                    stage.arguments[name], stage.arguments[name])

        if stage_json['type'] == 'Train':
            stage.arguments['epochs'] = 0
            break
    new_w = flowws.Workflow(stages)
    scope = new_w.run()
    model = scope['model']
    model.set_weights(weights)

    return LoadResult(model, scope)

@flowws.add_stage_arguments
class BackmapCoarseGrainViewer(flowws.Stage):
    """Visualize the results of backmapped coarse-grainings.

    This module creates a plato scene to visualize a coarse-grained
    system and the predicted fine-scale coordinates of a trained
    model. It loads a gtar file containing a trained model and flowws
    workflow definition (using `PDBCoarseGrained`, for example) and
    enables visualization of the coarse- and fine-grained coordinates.

    Workflows run on different systems will likely require use of the
    `path_substitutions` argument for their data-loading modules.

    """

    ARGS = [
        Arg('filename', '-f', str,
            help='Saved model to open'),
        Arg('path_substitutions', '-p', [(str, str)],
            help='Paths to replace in stage descriptions'),
        Arg('additive_rendering', '-a', bool, True,
            help='If True, use additive rendering'),
        Arg('fast_antialiasing', None, bool, False,
            help='Use Fast Approximate Antialiasing (FXAA)'),
        Arg('ambient_occlusion', None, bool, False,
            help='Use Screen Space Ambient Occlusion (SSAO)'),
        Arg('diameter_scale', '-d', float, 1.,
            help='Size scale for atomic diameters'),
        Arg('true_coords', None, bool, 1,
            help='Show atoms at their true coordinates instead of predicted coordinates'),
        Arg('show_amino_acids', None, bool, 1,
            help='Show amino acid center of mass beads'),
    ]

    def run(self, scope, storage):
        import plato, plato.draw as draw

        loaded = load_model(
            self.arguments['filename'], tuple(self.arguments['path_substitutions']))
        self.model = loaded.model
        self.loaded_scope = loaded.scope

        features = {}
        if self.arguments['additive_rendering']:
            features['additive_rendering'] = True

        prot_prim = draw.Spheres()
        atom_prim = draw.Spheres(diameters=.5*self.arguments['diameter_scale'])
        lines = draw.Lines(widths=.25, cap_mode=1)
        self.scene = draw.Scene([prot_prim, lines, atom_prim], zoom=4, features=features)

        named_type_colors = [
            elt_colors.get(name.split(';')[-1], (.5, .5, .5, 1))
            for name in self.loaded_scope['child_type_names']]
        named_type_colors = np.array(named_type_colors)

        named_type_radii = [
            elt_radii.get(name.split(';')[-1], .125)
            for name in self.loaded_scope['child_type_names']]
        named_type_diameters = np.array(named_type_radii)*2*self.arguments['diameter_scale']

        max_types = len(self.loaded_scope['type_names'])

        sample_batch = next(self.loaded_scope['train_generator'])
        ((sx, st, sct), sy) = sample_batch
        types = np.argmax(st[0, max_types:], axis=-1)
        dy = sy[0]
        diff, sum_ = st[0, :, :max_types], st[0, :, max_types:]
        types = np.argmax(sum_ - diff, axis=-1)
        ti, tj = np.argmax(sum_ - diff, axis=-1), np.argmax(sum_ + diff, axis=-1)

        cfilt = sct[0] != 0
        filt = tj != 0

        pred = self.model.predict((sx, st, sct))[0][cfilt]*self.loaded_scope['y_scale']

        coords = np.concatenate([sx[0][filt]*self.loaded_scope['x_scale'], [(0, 0, 0)]], axis=0)
        prot_prim.positions = coords
        if not self.arguments['show_amino_acids']:
            prot_prim.diameters = 0
        colors = np.ones((len(coords), 4))
        colors[:-1, :3] = plato.cmap.cubeellipse_intensity(tj[filt].astype(np.float32))
        colors[-1:, :3] = plato.cmap.cubeellipse_intensity(ti[:1].astype(np.float32))
        prot_prim.colors = colors
        if self.arguments['true_coords']:
            atom_prim.positions = sy[0][cfilt]*self.loaded_scope['y_scale']
        else:
            atom_prim.positions = pred
        atom_prim.colors = named_type_colors[sct[0][cfilt]]
        atom_prim.diameters = named_type_diameters[sct[0][cfilt]]
        lines.start_points = pred
        lines.end_points = sy[0][cfilt]*self.loaded_scope['y_scale']
        lines.colors = named_type_colors[sct[0][cfilt]]

        scope.setdefault('visuals', []).append(self)

        self.gui_actions = [
            ('Visualize next molecule', self._next_molecule),
        ]

    def _next_molecule(self, scope, storage):
        if scope.get('rerun_callback', None) is not None:
            scope['rerun_callback']()

    def draw_plato(self):
        return self.scene



from .internal import GradientLayer, MomentumNormalization, \
    NeighborhoodReduction, \
    PairwiseValueNormalization, PairwiseVectorDifference, \
    PairwiseVectorDifferenceSum, VectorAttention

import flowws
from flowws import Argument as Arg
import numpy as np
import tensorflow as tf
from tensorflow import keras

LAMBDA_ACTIVATIONS = {
    'log1pswish': lambda x: tf.math.log1p(tf.nn.swish(x)),
    'sin': tf.sin,
}

NORMALIZATION_LAYERS = {
    'layer': lambda _: keras.layers.LayerNormalization(),
    'layer_all': lambda rank: keras.layers.LayerNormalization(axis=[-i - 1 for i in range(rank + 1)]),
    'momentum': lambda _: MomentumNormalization(),
    'pairwise': lambda _: PairwiseValueNormalization(),
}

NORMALIZATION_LAYER_DOC = ' (any of {})'.format(','.join(NORMALIZATION_LAYERS))

@flowws.add_stage_arguments
class MoleculeForceRegression(flowws.Stage):
    """Build a geometric attention network for the molecular force regression task.

    This module specifies the architecture of a network to calculate
    atomic forces given the coordinates and types of atoms in a
    molecule. Conservative forces are computed by calculating the
    gradient of a scalar.

    """

    ARGS = [
        Arg('rank', None, int, 2,
            help='Degree of correlations (n-vectors) to consider'),
        Arg('n_dim', '-n', int, 32,
            help='Working dimensionality of point representations'),
        Arg('dilation', None, float, 2,
            help='Working dimension dilation factor for MLP components'),
        Arg('merge_fun', '-m', str, 'concat',
            help='Method to merge point representations'),
        Arg('join_fun', '-j', str, 'concat',
            help='Method to join invariant and point representations'),
        Arg('dropout', '-d', float, 0,
            help='Dropout rate to use, if any'),
        Arg('mlp_layers', None, int, 1,
            help='Number of hidden layers for score/value MLPs'),
        Arg('n_blocks', '-b', int, 2,
            help='Number of deep blocks to use'),
        Arg('block_nonlinearity', None, bool, True,
            help='If True, add a nonlinearity to the end of each block'),
        Arg('residual', '-r', bool, True,
            help='If True, use residual connections within blocks'),
        Arg('activation', '-a', str, 'swish',
            help='Activation function to use inside the network'),
        Arg('final_activation', None, str, 'swish',
            help='Final activation function to use within the network'),
        Arg('score_normalization', None, [str], [],
            help=('Normalizations to apply to score (attention) function' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('value_normalization', None, [str], [],
            help=('Normalizations to apply to value function' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('normalize_distances', None, str,
            help='Method to use to normalize pairwise distances'),
    ]

    def run(self, scope, storage):
        rank = self.arguments['rank']

        if self.arguments['activation'] in LAMBDA_ACTIVATIONS:
            activation_layer = lambda: keras.layers.Lambda(
                LAMBDA_ACTIVATIONS[self.arguments['activation']])
        else:
            activation_layer = lambda: keras.layers.Activation(
                self.arguments['activation'])

        if self.arguments['final_activation'] in LAMBDA_ACTIVATIONS:
            final_activation_layer = lambda: keras.layers.Lambda(
                LAMBDA_ACTIVATIONS[self.arguments['final_activation']])
        else:
            final_activation_layer = lambda: keras.layers.Activation(
                self.arguments['final_activation'])

        n_dim = self.arguments['n_dim']
        dilation_dim = int(np.round(n_dim*self.arguments['dilation']))

        def make_scorefun():
            layers = []

            for _ in range(self.arguments['mlp_layers']):
                layers.append(keras.layers.Dense(dilation_dim))

                for name in self.arguments['score_normalization']:
                    layers.append(NORMALIZATION_LAYERS[name](rank))

                layers.append(activation_layer())

                if self.arguments.get('dropout', 0):
                    layers.append(keras.layers.Dropout(self.arguments['dropout']))

            layers.append(keras.layers.Dense(1))
            return keras.models.Sequential(layers)

        def make_valuefun():
            layers = []

            for _ in range(self.arguments['mlp_layers']):
                layers.append(keras.layers.Dense(dilation_dim))

                for name in self.arguments['value_normalization']:
                    layers.append(NORMALIZATION_LAYERS[name](rank))

                layers.append(activation_layer())

                if self.arguments.get('dropout', 0):
                    layers.append(keras.layers.Dropout(self.arguments['dropout']))

            layers.append(keras.layers.Dense(n_dim))
            return keras.models.Sequential(layers)

        def make_block(last):
            residual_in = last
            last = VectorAttention(
                make_scorefun(), make_valuefun(), False, rank=rank,
                join_fun=self.arguments['join_fun'],
                merge_fun=self.arguments['merge_fun'])([delta_x, last])

            if self.arguments['block_nonlinearity']:
                last = make_valuefun()(last)

            if self.arguments['residual']:
                last = last + residual_in

            return last

        x_in = keras.layers.Input((scope['neighborhood_size'], 3))
        v_in = keras.layers.Input((scope['neighborhood_size'], scope['num_types']))

        delta_x = PairwiseVectorDifference()(x_in)
        delta_v = PairwiseVectorDifferenceSum()(v_in)

        if 'normalize_distances' in self.arguments:
            mode = self.arguments['normalize_distances']

            if mode == 'momentum':
                delta_x = MomentumNormalization()(delta_x)
            else:
                raise NotImplementedError(mode)

        last = keras.layers.Dense(n_dim)(delta_v)
        for _ in range(self.arguments['n_blocks']):
            last = make_block(last)

        (last, ivs, att) = VectorAttention(
            make_scorefun(), make_valuefun(), True, name='final_attention',
            rank=rank,
            join_fun=self.arguments['join_fun'],
            merge_fun=self.arguments['merge_fun'])(
            [delta_x, last], return_invariants=True, return_attention=True)

        last = keras.layers.Dense(dilation_dim, name='final_mlp')(last)
        last = final_activation_layer()(last)
        last = NeighborhoodReduction()(last)
        last = keras.layers.Dense(1, name='energy_projection', use_bias=False)(last)
        last = GradientLayer()((last, x_in))

        scope['input_symbol'] = [x_in, v_in]
        scope['output'] = last
        scope['loss'] = 'mse'
        scope['attention_model'] = keras.models.Model([x_in, v_in], att)
        scope['invariant_model'] = keras.models.Model([x_in, v_in], ivs)




from .models.internal import \
    vecvec, vecvec_invariants, vecvec_covariants, \
    bivecvec, bivecvec_invariants, bivecvec_covariants, \
    trivecvec, trivecvec_invariants, trivecvec_covariants



import collections
import random

import flowws
from flowws import Argument as Arg
import freud
import numpy as np

from .internal import ScaledMSE, ScaledMAE

CoarseSystem = collections.namedtuple('CoarseSystem',
   ['box', 'nlist', 'positions', 'types', 'type_names',
    'child_positions', 'child_types', 'child_type_names'])

def coarse_grain(record, num_neighbors=4, x_scale=1.):
    positions = []
    types = []
    child_positions = []
    child_types = []
    index_groups = np.split(np.arange(len(record.residue_ids)),
                            np.unique(record.residue_ids, True)[1])[1:]
    for group in index_groups:
        group_child_positions = record.positions[group]/x_scale
        group_child_types = record.types[group]
        center_of_mass = np.mean(group_child_positions, axis=0)
        positions.append(center_of_mass)
        types.append(record.residue_types[group[0]])
        child_positions.append(group_child_positions)
        child_types.append(group_child_types)

    positions = np.array(positions, dtype=np.float32)
    types = np.array(types, dtype=np.uint32)
    box = np.array(record.box, dtype=np.float32)/x_scale

    aabb = freud.locality.AABBQuery(box, positions)
    query = aabb.query(positions, dict(num_neighbors=num_neighbors, exclude_ii=True, mode='nearest'))
    nlist = query.toNeighborList()

    return CoarseSystem(
        box, nlist, positions, types, record.residue_type_names,
        child_positions, child_types, record.type_names)

def loop_neighborhood_environments(
        rec, neighborhood_size, seed=13, fraction_range=(0, 2.),
        randomly_rotate=False):

    rand = np.random.default_rng(seed)
    index_i = rec.nlist.query_point_indices
    index_j = rec.nlist.point_indices

    shuffle_indices = np.arange(len(rec.positions))
    fraction_assignments = np.linspace(0, 1, len(rec.positions), endpoint=False)
    rand.shuffle(fraction_assignments)
    filt = np.logical_and(fraction_range[0] <= fraction_assignments,
                          fraction_assignments < fraction_range[1])
    if not np.sum(filt):
        raise ValueError(
            'No particles found for fraction_range: {}'.format(fraction_range))
    shuffle_indices = shuffle_indices[filt]

    while True:
        rand.shuffle(shuffle_indices)

        for i in shuffle_indices:
            bond_start = rec.nlist.find_first_index(i)
            bond_stop = rec.nlist.find_first_index(i + 1)
            bonds = slice(bond_start, bond_stop)
            r0 = rec.positions[index_i[bond_start]]
            rij = (rec.positions[index_j[bonds]] - rec.positions[index_i[bonds]])

            types_j = rec.types[index_j[bonds]]
            types_i = rec.types[index_i[bonds]]

            rchildren = rec.child_positions[i] - r0
            tchildren = rec.child_types[i]

            if randomly_rotate:
                import rowan
                q = rowan.random.rand(1)[None]
                rij = rowan.rotate(q, rij)
                rchildren = rowan.rotate(q, rchildren)

            yield rij, types_i, types_j, rchildren, tchildren

def randomly_loop_iter(xs, seed):
    rand = random.Random(seed)
    xs = list(xs)
    while True:
        rand.shuffle(xs)
        yield from xs

def make_batches(cg_records, batch_size, neighborhood_size,
                 max_atoms, max_types, global_type_remaps, y_scale=1.,
                 fraction_range=(0, 2.), seed=13, randomly_rotate=False):
    rand = random.Random(seed)
    name_iter = randomly_loop_iter(sorted(cg_records), rand.randint(0, 2**32))

    iterators = {}
    for (name, rec) in sorted(cg_records.items()):
        iterators[name] = loop_neighborhood_environments(
            rec, neighborhood_size, seed + 1, fraction_range,
            randomly_rotate=randomly_rotate)

    while True:
        cg_rij = np.zeros((batch_size, neighborhood_size, 3), dtype=np.float32)
        cg_tij = np.zeros((batch_size, neighborhood_size, 2*max_types), dtype=np.float32)
        fg_tchild = np.zeros((batch_size, max_atoms), dtype=np.uint32)
        fg_rchild = np.zeros((batch_size, max_atoms, 3), dtype=np.float32)

        for batch_i in range(batch_size):
            name = next(name_iter)
            (res_type_remap, atom_type_remap) = global_type_remaps[name]
            (rij, types_i, types_j, rchildren, tchildren) = next(iterators[name])

            types_i, types_j = res_type_remap[types_i], res_type_remap[types_j]
            types_i = np.eye(max_types)[types_i]
            types_j = np.eye(max_types)[types_j]

            cg_rij[batch_i, :len(rij)] = rij
            cg_tij[batch_i, :len(rij), :max_types] = types_j + types_i
            cg_tij[batch_i, :len(rij), max_types:] = types_j - types_i
            fg_tchild[batch_i, :len(rchildren)] = atom_type_remap[tchildren]
            fg_rchild[batch_i, :len(rchildren)] = rchildren/y_scale

        yield (cg_rij, cg_tij, fg_tchild), fg_rchild

@flowws.add_stage_arguments
class PDBCoarseGrained(flowws.Stage):
    """Generate data for backmapping coarse-graining of structures from the PDB.

    This module calculates coarse-grained versions of PDB records,
    loaded using `PDBCache`. Generators are produced for model
    training.

    """

    ARGS = [
        Arg('neighborhood_size', '-n', int,
           help='Neighborhood size (number of input amino acid coordinates) to use'),
        Arg('batch_size', '-b', int, 32,
           help='Batch size to use'),
        Arg('seed', '-s', int, 14,
           help='Random seed to use'),
        Arg('validation_fraction', '-v', float, .3,
           help='Fraction of record names to be assigned to validation set'),
        Arg('test_fraction', '-t', float,
           help='Fraction of record names to be assigned to validation set'),
        Arg('x_scale', '-x', float, 64.,
           help='Factor by which to decrease residue length scales'),
        Arg('y_scale', '-y', float, 8.,
           help='Factor by which to decrease atomic length scales'),
        Arg('randomly_rotate', '-r', bool, False,
            help='If True, randomly rotate environments'),
    ]

    def run(self, scope, storage):

        all_records = scope['pdb_records']
        coarse_records = {}
        skipped_records = []
        for (name, rec) in all_records.items():
            coarse = coarse_grain(
                rec, self.arguments['neighborhood_size'], self.arguments['x_scale'])

            if any(np.max(np.bincount(ts)) > 1 for ts in coarse.child_types):
                skipped_records.append((name, 'duplicate child types'))
                continue
            if len(coarse.positions) <= self.arguments['neighborhood_size']:
                skipped_records.append((name, 'too few positions'))
                continue
            coarse_records[name] = coarse

        scope['coarse_records'] = coarse_records
        scope['skipped_records'] = skipped_records
        print('{} final records'.format(len(coarse_records)))
        print('{} skipped records'.format(len(skipped_records)))

        max_atoms = 0
        all_residue_types, all_atom_types = set(), set()
        for rec in coarse_records.values():
            all_residue_types.update(rec.type_names)
            all_atom_types.update(rec.child_type_names)
            max_atoms = max(max_atoms, max(len(v) for v in rec.child_positions))

        all_residue_types = ['NORES'] + list(sorted(all_residue_types))
        residue_type_map = {name: i for (i, name) in enumerate(all_residue_types)}
        all_atom_types = ['NOATM'] + list(sorted(all_atom_types))
        atom_type_map = {name: i for (i, name) in enumerate(all_atom_types)}

        global_type_remaps = {}
        for (name, rec) in coarse_records.items():
            res_type_remap = [residue_type_map[name] for name in rec.type_names]
            res_type_remap = np.array(res_type_remap, dtype=np.uint32)
            atom_type_remap = [atom_type_map[name] for name in rec.child_type_names]
            atom_type_remap = np.array(atom_type_remap, dtype=np.uint32)
            global_type_remaps[name] = (res_type_remap, atom_type_remap)

        print('Max number of atoms in a residue:', max_atoms)

        scaled_mse = ScaledMSE(self.arguments['y_scale'])
        scaled_mae = ScaledMAE(self.arguments['y_scale'])
        y_scale = self.arguments['y_scale']/self.arguments['x_scale']

        ranges, labels = [0], []
        if self.arguments['validation_fraction']:
            ranges.append(self.arguments['validation_fraction'])
            labels.append('validation')

        if 'test_fraction' in self.arguments:
            ranges.append(self.arguments['test_fraction'])
            labels.append('test')

        ranges.append(2.)
        labels.append('train')

        cumulative_ranges = np.cumsum(ranges)
        label_ranges = {name: (start, stop) for (name, start, stop)
                        in zip(labels, cumulative_ranges[:-1], cumulative_ranges[1:])}

        for (name, fraction_range) in label_ranges.items():
            scope['{}_generator'.format(name)] = make_batches(
                coarse_records, self.arguments['batch_size'], self.arguments['neighborhood_size'],
                max_atoms, len(all_residue_types), global_type_remaps, y_scale,
                fraction_range, self.arguments['seed'],
                randomly_rotate=self.arguments['randomly_rotate'])

        if 'validation_generator' not in scope:
            scope['validation_generator'] = scope['train_generator']

        scope['x_scale'] = self.arguments['x_scale']
        scope['y_scale'] = self.arguments['y_scale']
        scope['type_names'] = all_residue_types
        scope['type_name_map'] = residue_type_map
        scope['child_type_names'] = all_atom_types
        scope['child_type_name_map'] = atom_type_map
        scope.setdefault('metrics', []).extend([scaled_mse, scaled_mae])



#!/usr/bin/env python

import os
from setuptools import setup

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

version_fname = os.path.join(THIS_DIR, 'flowws_keras_geometry', 'version.py')
with open(version_fname) as version_file:
    exec(version_file.read())

readme_fname = os.path.join(THIS_DIR, 'README.md')
with open(readme_fname) as readme_file:
    long_description = readme_file.read()

entry_points = set()
flowws_modules = []
package_names = ['flowws_keras_geometry']

def add_subpkg(subpkg, module_names):
    package_names.append('flowws_keras_geometry.{}'.format(subpkg))
    for name in module_names:
        if name not in entry_points:
            flowws_modules.append('{0} = flowws_keras_geometry.{1}.{0}:{0}'.format(name, subpkg))
            entry_points.add(name)
        flowws_modules.append(
            'flowws_keras_geometry.{1}.{0} = flowws_keras_geometry.{1}.{0}:{0}'.format(name, subpkg))

module_names = [
]
for name in module_names:
    if name not in entry_points:
        flowws_modules.append('{0} = flowws_keras_geometry.{0}:{0}'.format(name))
        entry_points.add(name)
    flowws_modules.append(
        'flowws_keras_geometry.{0} = flowws_keras_geometry.{0}:{0}'.format(name))

subpkg = 'data'
module_names = [
    'MD17',
    'PDBCache',
    'PDBCoarseGrained',
    'PyriodicDataset',
    'RMD17',
]
add_subpkg(subpkg, module_names)

subpkg = 'models'
module_names = [
    'CrystalStructureClassification',
    'MoleculeForceRegression',
    'PDBInverseCoarseGrain',
    'PDBInverseCoarseGrainTransformer',
]
add_subpkg(subpkg, module_names)

subpkg = 'viz'
module_names = [
    'BackmapCoarseGrainViewer',
    'MoleculeAttentionViewer',
    'ParticleAttentionViewer',
]
add_subpkg(subpkg, module_names)

setup(name='flowws-keras-geometry',
      author='Matthew Spellings',
      author_email='mspells@vectorinstitute.ai',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
      ],
      description='Stage-based scientific workflows for deep learning experiments with geometric algebra',
      entry_points={
          'flowws_modules': flowws_modules,
      },
      extras_require={},
      install_requires=[
          'flowws',
          'flowws-keras-experimental',
      ],
      license='MIT',
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=package_names,
      python_requires='>=3',
      version=__version__
      )




# -- Path setup --------------------------------------------------------------

import sys
from unittest.mock import MagicMock
import os

with open('../../flowws_keras_geometry/version.py') as version_file:
    exec(version_file.read())
sys.path.insert(0, os.path.abspath('../..'))

class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()

    @staticmethod
    def __setitem__(key, value):
        pass

autodoc_mock_imports = [
    'freud',
    'keras_gtar',
    'pyriodic',
]

sys.modules.update((mod_name, Mock()) for mod_name in autodoc_mock_imports)

# -- Project information -----------------------------------------------------

project = 'flowws-keras-geometry'
copyright = '2021, Matthew Spellings'
author = 'Matthew Spellings'
version = __version__
release = version

# -- General configuration ---------------------------------------------------

extensions = [
    'm2r2',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



import os
import subprocess

import flowws
from flowws import Argument as Arg
import numpy as np
from tensorflow import keras

from .internal import ScaledMSE, ScaledMAE
from .MD17 import MD17

@flowws.add_stage_arguments
class RMD17(MD17):
    """Load data from the RMD17 dataset for molecular force regression.

    This module downloads the entire RMD17 dataset. It randomly
    selects molecule snapshots for the specified molecules to produce
    training, testing, and validation sets.

    """

    def _download(self, name):
        fname = 'rmd17_{}.npz'.format(name.split('_')[0])

        if fname not in os.listdir(self.arguments['cache_dir']):
            url = 'https://ndownloader.figshare.com/files/23950376'
            output_name = os.path.join(self.arguments['cache_dir'], 'RMD17.tar.bz2')
            command = ['wget', '-c', '-O', output_name, url]
            subprocess.check_call(command)

            command = ['tar', 'xf', output_name, '-C', self.arguments['cache_dir'], '--strip', '2']
            subprocess.check_call(command)

        return os.path.join(self.arguments['cache_dir'], fname)

    @staticmethod
    def get_encoding(data, max_atoms, type_map, indices=None, energy_conversion=1.):
        coords = data['coords']
        forces = data['forces']*energy_conversion
        types = np.zeros(max_atoms, dtype=np.uint32)
        types[:coords.shape[1]] = [type_map[t] for t in data['nuclear_charges']]

        if indices is not None:
            coords = coords[indices]
            forces = forces[indices]

        rs = np.zeros((len(coords), max_atoms, 3))
        rs[:, :coords.shape[1], :] = coords

        Fs = np.zeros((len(coords), max_atoms, 3))
        Fs[:, :coords.shape[1], :] = forces

        types_onehot = np.eye(len(type_map))[types]
        types_onehot = np.tile(types_onehot[np.newaxis, ...], (len(coords), 1, 1))
        return (rs, types_onehot), Fs

    @staticmethod
    def get_trajectory_size_types(data):
        (frames, size, _) = data['coords'].shape
        return (frames, size, data['nuclear_charges'])



import collections
import os
import subprocess

import flowws
from flowws import Argument as Arg
import numpy as np

PDBRecord = collections.namedtuple('PDBRecord',
    ['box', 'positions', 'types', 'type_names', 'residue_ids', 'residue_types', 'residue_type_names'])
#            (N, 3),       (N,)     (N_types,)     (N,)          (N_res)           (N_res,)

def parse_pdb(filename, ignore_hydrogens=False):
    cmd = ['gunzip', '-ckf', filename]

    positions = []
    types = []
    residue_ids = []
    residue_types = []

    type_name_map = collections.defaultdict(lambda: len(type_name_map))
    residue_type_name_map = collections.defaultdict(lambda: len(residue_type_name_map))
    res_id = -1

    for line in subprocess.check_output(cmd).decode().splitlines():
        if line.startswith('MODEL'):
            last_residue = (None, None)

        if line.startswith('TER'):
            last_residue = (None, None)

        if not line.startswith('ATOM'):
            continue

        # skip non-blank alternate locations
        if line[16] != ' ':
            continue

        coords = tuple(map(float, (line[start:end] for (start, end) in
                                   [(30, 38), (38, 46), (46, 54)])))

        # special atomic type, like CA/CB/etc
        atype = line[12:16].strip()
        # actual element type
        etype = line[76:78].strip()

        res_id_str = line[22:26].strip()
        res_type = line[17:20].strip()
        current_residue = (res_id_str, res_type)

        res_id += current_residue != last_residue
        last_residue = current_residue

        if (etype == 'H' or etype == 'D') and ignore_hydrogens:
            continue

        positions.append(coords)
        types.append(type_name_map['{};{}'.format(atype, etype)])
        residue_ids.append(res_id)
        residue_types.append(residue_type_name_map[res_type])

    box = (2*np.max(positions, axis=0) - 2*np.min(positions, axis=0)).tolist()
    box.extend([0, 0, 0])

    positions = np.array(positions, dtype=np.float32)
    positions -= np.mean(positions, axis=0, keepdims=True)

    types = np.array(types, dtype=np.uint32)
    type_names = [v for (_, v) in sorted((v, i) for (i, v) in type_name_map.items())]

    residue_ids = np.array(residue_ids, dtype=np.int32)
    residue_types = np.array(residue_types, dtype=np.uint32)
    residue_type_names = [v for (_, v) in sorted((v, i) for (i, v) in residue_type_name_map.items())]

    return PDBRecord(box, positions, types, type_names, residue_ids, residue_types, residue_type_names)

@flowws.add_stage_arguments
class PDBCache(flowws.Stage):
    """Load PDB records and cache them in a local directory.

    This module stores and reads atomic coordinates from a set of
    specified PDB records. Files are cached in a local directory to
    avoid repeated downloads.

    """

    ARGS = [
        Arg('cache_directory', None, str, '/tmp',
           help='Cache directory to use for downloaded PDB records'),
        Arg('records', '-r', [str],
           help='Record IDs to use for training (or file://path/to/filename.txt'
            ' to read lines from a file)'),
        Arg('failure_mode', '-f', str, 'default',
           help='Behavior when parsing/fetching a file fails (default/ignore)'),
        Arg('ignore_hydrogens', None, bool, True,
           help='If True, ignore hydrogens when building systems'),
    ]

    def _maybe_get_pdbfile(self, pdb_id):
        fname = '{}.pdb1.gz'.format(pdb_id)
        target = os.path.join(self.arguments['cache_directory'], fname)

        if not os.path.exists(target):
            url = 'https://files.rcsb.org/download/{}'.format(fname)
            cmd = ['wget', url, '-O', target]
            subprocess.check_call(cmd)

        return target

    def run(self, scope, storage):
        record_names = []
        for name in self.arguments['records']:
            if name.startswith('file://'):
                with storage.open(name[len('file://'):], 'r') as f:
                    for line in f.splitlines():
                        record_names.append(line.strip())
            else:
                record_names.append(name)

        parse_kwargs = dict(ignore_hydrogens=self.arguments['ignore_hydrogens'])

        all_records = {}
        for pdb_id in record_names:
            if self.arguments['failure_mode'] in ('ignore', 'print'):
                try:
                    fname = self._maybe_get_pdbfile(pdb_id)
                    all_records[pdb_id] = parse_pdb(fname, **parse_kwargs)
                except Exception as e:
                    if self.arguments['failure_mode'] == 'print':
                        print(pdb_id, e)
                    pass
            else:
                fname = self._maybe_get_pdbfile(pdb_id)
                all_records[pdb_id] = parse_pdb(fname, **parse_kwargs)

        scope['pdb_records'] = all_records
